//Contiguous Derived Datatype: C Example

// Blocking Send does not allow rank = 0 send to itself. Rank = 0 send to itself, program will be blocked at the send sentence and never reach the recieve sentence, which leads a deadlocking. 
#include "mpi.h" 
#include "stdio.h"
#include "stdlib.h" 
#define SIZE 4
typedef struct {
  double u;
  double v;
} TESTVAR;
 
int main(argc,argv) 
  int argc; 
  char *argv[ ];  { 
    int numtasks, rank, source=0, dest, tag=1, i,j; 
    TESTVAR **a; 
    TESTVAR **b, **c; 
    MPI_Status stat; 
    MPI_Datatype structype, oldtypes[3], rowtype, columntype;
    int blockcounts[3];
    MPI_Aint offsets[3];

    MPI_Init(&argc,&argv); 
    MPI_Comm_rank(MPI_COMM_WORLD, &rank); 
    MPI_Comm_size(MPI_COMM_WORLD, &numtasks);

    a = (TESTVAR**) calloc(SIZE, sizeof(TESTVAR*));
    b = (TESTVAR**) calloc(SIZE, sizeof(TESTVAR*));
    c = (TESTVAR**) calloc(SIZE, sizeof(TESTVAR*));
    for(i = 0; i < SIZE; i++) {
      a[i] = (TESTVAR*) calloc(SIZE, sizeof(TESTVAR));
      b[i] = (TESTVAR*) calloc(SIZE, sizeof(TESTVAR));
      c[i] = (TESTVAR*) calloc(SIZE, sizeof(TESTVAR));
    }
    if(!rank) {
      for(i=0; i<SIZE; i++) {
	for(j=0; j<SIZE; j++) {
	  a[i][j].u = 1.0*(i*SIZE + j + 1);
	  a[i][j].v = (i*SIZE + j + 1)*10.;
	  b[i][j].u = -5.;
	  b[i][j].v = -5.;
	}
      }

      for(i=0; i<SIZE; i++) {
	for(j=0; j<SIZE; j++) {
	  printf("rank = %d, %f  %f  \n", rank, a[i][j].u, a[i][j].v);
	}
      }
    } else {}


    TESTVAR tv;
    offsets[0] = (MPI_Aint)(&tv.u) - (MPI_Aint)(&tv);
    oldtypes[0] = MPI_DOUBLE;
    blockcounts[0] = 1;
    offsets[1] = (MPI_Aint)(&tv.v) - (MPI_Aint)(&tv);
    oldtypes[1] = MPI_DOUBLE;
    blockcounts[1] = 1;
    offsets[2] = sizeof(TESTVAR);
    oldtypes[2] = MPI_UB;
    blockcounts[2] = 1;

//    for(i = 0; i <= 2; i++) printf("offset[%d] = %d\n", i, offsets[i]);

    MPI_Type_struct(3, blockcounts, offsets, oldtypes, &structype); 
    MPI_Type_commit(&structype);

    MPI_Type_contiguous(SIZE, structype, &rowtype);
    MPI_Type_commit(&rowtype);
    MPI_Type_vector(SIZE, 1, SIZE, structype, &columntype);
    MPI_Type_commit(&columntype);
  
    if (numtasks == SIZE) { 
      if (rank == 0) { 
	for (i=1; i < numtasks; i++) { 
	  MPI_Send(&a[i][0], 1, rowtype, i, i, MPI_COMM_WORLD); 
          MPI_Send(&a[0][i], 1, columntype, i, 10*i, MPI_COMM_WORLD);
        }
      } else {} 

      if(rank) {

        MPI_Recv(&b[rank][0], 1, rowtype, source, rank, MPI_COMM_WORLD, &stat);
        printf("rank= %d  b.u= %3.1f %3.1f %3.1f %3.1f\n", rank,b[rank][0].u,b[rank][1].u,b[rank][2].u,b[rank][3].u);
    	printf("rank= %d  b.v= %3.1f %3.1f %3.1f %3.1f\n", rank,b[rank][0].v,b[rank][1].v,b[rank][2].v,b[rank][3].v);

        MPI_Recv(&c[0][rank], 1, columntype, source, rank*10, MPI_COMM_WORLD, &stat);
        printf("rank= %d  c.u= %3.1f %3.1f %3.1f %3.1f\n", rank,c[0][rank].u,c[1][rank].u,c[2][rank].u,c[3][rank].u);
        printf("rank= %d  c.v= %3.1f %3.1f %3.1f %3.1f\n", rank,c[0][rank].v,c[1][rank].v,c[2][rank].v,c[3][rank].v);


      } else {}
    } 
    else 
      printf("Must specify %d processors. Terminating.\n",SIZE); 

    MPI_Finalize(); 
  } 
